package NW01;

use Exporter;
@ISA = (Exporter);
@EXPORT = qw(nw);
    
sub nw{

# Set query and database sequences
my $q_seq = shift;
my $d_seq = shift;

# Gap and gap extension penalties
my $gap = -5;
my $ext = -1;

# Convert sequences into arrays
my @q_aas = split(//,$q_seq);
my @d_aas = split(//,$d_seq);

# Hash for conversion of amino acid to number
my $all_aa_str = "CSTPAGNDEQHRKMILVFYW";
my @all_aas = split(//,$all_aa_str);
my %an_cnvd;
my $num = 0;
foreach (@all_aas){
    $an_cnvd{$_} = $num;
    $num++;
}

# Convert amino acid alphapet to number
$_ = $an_cnvd{$_} foreach (@q_aas);
$_ = $an_cnvd{$_} foreach (@d_aas);

# 2D array for score matrix
my @score_matrix = ( #BLOSUM62
    [9,-1,-1,-3,0,-3,-3,-3,-4,-3,-3,-3,-3,-1,-1,-1,-1,-2,-2,-2], #C
    [-1,4,1,-1,1,0,1,0,0,0,-1,-1,0,-1,-2,-2,-2,-2,-2,-3],  #S
    [-1,1,4,1,-1,1,0,1,0,0,0,-1,0,-1,-2,-2,-2,-2,-2,-3],  #T
    [-3,-1,1,7,-1,-2,-1,-1,-1,-1,-2,-2,-1,-2,-3,-3,-2,-4,-3,-4],   #P
    [0,1,-1,-1,4,0,-1,-2,-1,-1,-2,-1,-1,-1,-1,-1,-2,-2,-2,-3],    #A
    [-3,0,1,-2,0,6,-2,-1,-2,-2,-2,-2,-2,-3,-4,-4,0,-3,-3,-2],  #G
    [-3,1,0,-2,-2,0,6,1,0,0,-1,0,0,-2,-3,-3,-3,-3,-2,-4],  #N
    [-3,0,1,-1,-2,-1,1,6,2,0,-1,-2,-1,-3,-3,-4,-3,-3,-3,-4], #D
    [-4,0,0,-1,-1,-2,0,2,5,2,0,0,1,-2,-3,-3,-3,-3,-2,-3], #E
    [-3,0,0,-1,-1,-2,0,0,2,5,0,1,1,0,-3,-2,-2,-3,-1,-2],    #Q
    [-3,-1,0,-2,-2,-2,1,1,0,0,8,0,-1,-2,-3,-3,-2,-1,2,-2],    #H
    [-3,-1,-1,-2,-1,-2,0,-2,0,1,0,5,2,-1,-3,-2,-3,-3,-2,-3],    #R
    [-3,0,0,-1,-1,-2,0,-1,1,1,-1,2,5,-1,-3,-2,-3,-3,-2,-3], #K
    [-1,-1,-1,-2,-1,-3,-2,-3,-2,0,-2,-1,-1,5,1,2,-2,0,-1,-1],  #M
    [-1,-2,-2,-3,-1,-4,-3,-3,-3,-3,-3,-3,-3,1,4,2,1,0,-1,-3], #I
    [-1,-2,-2,-3,-1,-4,-3,-4,-3,-2,-3,-2,-2,2,2,4,3,0,-1,-2],   #L
    [-1,-2,-2,-2,0,-3,-3,-3,-2,-2,-3,-3,-2,1,3,1,4,-1,-1,-3],    #V
    [-2,-2,-2,-4,-2,-3,-3,-3,-3,-3,-1,-3,-3,0,0,0,-1,6,3,1], #F
    [-2,-2,-2,-3,-2,-3,-2,-3,-2,-1,2,-2,-2,-1,-1,-1,-1,3,7,2],   #Y
    [-2,-3,-3,-4,-3,-2,-4,-4,-3,-2,-2,-3,-3,-1,-3,-2,-3,1,2,11]  #W
);


# Creat align matrix
my @align_matrix;
my $q_length = length($q_seq);
my $d_length = length($d_seq);

# Fill align matrix with "0"
for (0..$q_length){
    $align_matrix[$_] = [(0) x ($d_length+1)];
}

# Re-fill the first column/row with gap penalty
for (1..$q_length){
    $align_matrix[$_][0] = $_==1? $gap : $gap+($_-1)*$ext;
}
for (1..$d_length){
    $align_matrix[0][$_] = $_==1? $gap : $gap+($_-1)*$ext;
}

# ***** Creat Backtrack Matrix***** #

# Compute align/backtrack matrix
my $k = 0;
my $gap_value;
for (my $i=1; $i<=$q_length; $i++) {
    for (my $j=1; $j<=$d_length; $j++) {
        if ($k == 0){
            $gap_value = $gap;
        }else{
            $gap_value = $ext*$k;
        }
        my ($tmp_score, $k, $dir) =&max(
                $align_matrix[$i-1][$j-1]+$score_matrix[$q_aas[$i-1]][$d_aas[$j-1]],
                $align_matrix[$i][$j-1]+$gap_value,
                $align_matrix[$i-1][$j]+$gap_value,
                $k
        );
        $align_matrix[$i][$j] = $tmp_score;
        # ***** Add Values to Bctk matrix ***** #
    }
}

# Subroutine for finding max score
sub max{
    my($d,$u,$l,$k)=@_;
    if($d>$u){
        return ($d, 0, 0) if($d>$l);
        return ($l, $k+1, 2);
    }
    else{
        return ($u, $k+1, 1) if($u>$l);
        return ($l, $k+1, 2);    
    }
}

return $align_matrix[$q_length][$d_length];
# print "Score = $align_matrix[$q_length][$d_length]\n";

# ***** Backtracking routine ***** #

}

return 1;

__END__

# ***** Creat Backtrack Matrix***** #
# Creat backtrack matrix
my @backtrack_matrix;

# Fill backtrack matrix with "0"
for (0..$q_length){
    $backtrack_matrix[$_] = [(0) x ($d_length+1)];
}

# ***** Add Values to Bctk matrix ***** #
$backtrack_matrix[$i][$j] = $dir;

# ***** Backtracking routine ***** #
# Backtrack align_matrix
my $i = $q_length;
my $j = $d_length;
my @align1;
my @align2;

while ($i > 0 && $j > 0) {    
    # Score diagonal
    if ($backtrack_matrix[$i][$j] == 0) {
        push @align1, $q_aas[$i-1];
        push @align2, $d_aas[$j-1];
        $i--;
        $j--;
    }
    # Score up
    elsif ($backtrack_matrix[$i][$j] == 1){
        push @align1, "-";
        push @align2, $d_aas[$j-1];
        $j--;        
    }
    # Score left
    elsif ($backtrack_matrix[$i][$j] == 2) {
        push @align1, $q_aas[$i-1];
        push @align2, "-";
        $i--;
    }
}

# Query sequence longer
while ($i > 0) {
    push @align1, $q_aas[$i-1];
    push @align2, "-";
    $i--;
}
# Database sequence longer
while ($j > 0) {
    push @align1, "-";
    push @align2, $d_aas[$j-1];
    $j--;
}

# Reverse
@align1 = reverse @align1;
@align2 = reverse @align2;

# Convert number back to amino acid or "-"
foreach (@align1){
    $_ = $_ eq '-'?'-':$all_aas[$_];
}
foreach (@align2){
    $_ = $_ eq '-'?'-':$all_aas[$_];
}

# Print out result
print "$_"foreach @align1;
print "\n";
print "$_"foreach @align2;
